@@ -284,28 +284,45 @@ readbuf_close (ReadBuf *rb)
 	g_free (rb);
 }
 
-static GnomeDesktopItemType G_GNUC_CONST
+static GnomeDesktopItemType
 type_from_string (const char *type)
 {
-	if (type == NULL) {
+	if (!type)
 		return GNOME_DESKTOP_ITEM_TYPE_NULL;
-	} else if (strcmp (type, "Application") == 0) {
-		return GNOME_DESKTOP_ITEM_TYPE_APPLICATION;
-	} else if (strcmp (type, "Link") == 0) {
-		return GNOME_DESKTOP_ITEM_TYPE_LINK;
-	} else if (strcmp (type, "FSDevice") == 0) {
-		return GNOME_DESKTOP_ITEM_TYPE_FSDEVICE;
-	} else if (strcmp (type, "MimeType") == 0) {
-		return GNOME_DESKTOP_ITEM_TYPE_MIME_TYPE;
-	} else if (strcmp (type, "Directory") == 0) {
-		return GNOME_DESKTOP_ITEM_TYPE_DIRECTORY;
-	} else if (strcmp (type, "Service") == 0) {
-		return GNOME_DESKTOP_ITEM_TYPE_SERVICE;
-	} else if (strcmp (type, "ServiceType") == 0) {
-		return GNOME_DESKTOP_ITEM_TYPE_SERVICE_TYPE;
-	} else {
-		return GNOME_DESKTOP_ITEM_TYPE_OTHER;
+
+	switch (type [0]) {
+	case 'A':
+		if (!strcmp (type, "Application"))
+			return GNOME_DESKTOP_ITEM_TYPE_APPLICATION;
+		break;
+	case 'L':
+		if (!strcmp (type, "Link"))
+			return GNOME_DESKTOP_ITEM_TYPE_LINK;
+		break;
+	case 'F':
+		if (!strcmp (type, "FSDevice"))
+			return GNOME_DESKTOP_ITEM_TYPE_FSDEVICE;
+		break;
+	case 'M':
+		if (!strcmp (type, "MimeType"))
+			return GNOME_DESKTOP_ITEM_TYPE_MIME_TYPE;
+		break;
+	case 'D':
+		if (!strcmp (type, "Directory"))
+			return GNOME_DESKTOP_ITEM_TYPE_DIRECTORY;
+		break;
+	case 'S':
+		if (!strcmp (type, "Service"))
+			return GNOME_DESKTOP_ITEM_TYPE_SERVICE;
+
+		else if (!strcmp (type, "ServiceType"))
+			return GNOME_DESKTOP_ITEM_TYPE_SERVICE_TYPE;
+		break;
+	default:
+		break;
 	}
+
+	return GNOME_DESKTOP_ITEM_TYPE_OTHER;
 }
 
 /**
@@ -392,6 +409,11 @@ gnome_desktop_item_copy (const GnomeDesktopItem *item)
 	retval->location = g_strdup (item->location);
 	retval->mtime = item->mtime;
 
+	/* Languages */
+	retval->languages = g_list_copy (item->languages);
+	for (li = retval->languages; li != NULL; li = li->next)
+		li->data = g_strdup (li->data);	
+
 	/* Keys */
 	retval->keys = g_list_copy (item->keys);
 	for (li = retval->keys; li != NULL; li = li->next)
@@ -538,7 +560,7 @@ gnome_desktop_item_new_from_uri (const char *uri,
 	info = gnome_vfs_file_info_new ();
 
 	result = gnome_vfs_get_file_info (uri, info,
-					  GNOME_VFS_FILE_INFO_DEFAULT);
+					  GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
 	if (result != GNOME_VFS_OK) {
 		g_set_error (error,
 			     /* FIXME: better errors */
@@ -593,7 +615,7 @@ gnome_desktop_item_new_from_uri (const char *uri,
 		subfn = g_build_filename (uri, ".directory", NULL);
 		gnome_vfs_file_info_clear (info);
 		if (gnome_vfs_get_file_info (subfn, info,
-					     GNOME_VFS_FILE_INFO_DEFAULT) != GNOME_VFS_OK) {
+					     GNOME_VFS_FILE_INFO_FOLLOW_LINKS) != GNOME_VFS_OK) {
 			gnome_vfs_file_info_unref (info);
 			g_free (subfn);
 
@@ -1074,18 +1096,28 @@ stringify_files (GSList *args,
 
 	for (li = args; li != NULL; li = li->next) {
 		GnomeVFSURI *uri = li->data;
-		if (gnome_vfs_uri_is_local (uri)) {
-			const char *path;
+		if (!strcmp (gnome_vfs_uri_get_scheme (uri), "file")) {
+			char *path, *local_path;
 			char *escaped;
-			path = gnome_vfs_uri_get_path (uri);
-			g_string_append (str, sep);
-
-			escaped = escape_single_quotes (path,
-							in_single_quotes,
-							in_double_quotes);
-			g_string_append (str, escaped);
-			g_free (escaped);
+			
+			path = gnome_vfs_uri_to_string (uri, GNOME_VFS_URI_HIDE_NONE);
+			local_path = gnome_vfs_get_local_path_from_uri (path);
 
+			if (local_path != NULL) {
+				/* should never be null since we check for scheme
+				 * above, but you never know ...
+				 */
+				g_string_append (str, sep);
+	
+				escaped = escape_single_quotes (local_path,
+								in_single_quotes,
+								in_double_quotes);
+				g_string_append (str, escaped);
+				g_free (escaped);
+				g_free (local_path);
+			}
+			
+			g_free (path);
 			sep = " ";
 		}
 
@@ -1443,8 +1475,10 @@ static int
 ditem_execute (const GnomeDesktopItem *item,
 	       const char *exec,
 	       GList *file_list,
+               char **envp,
 	       gboolean launch_only_one,
 	       gboolean use_current_dir,
+		 gboolean append_uris,
 	       GError **error)
 {
 	char **real_argv;
@@ -1457,7 +1491,7 @@ ditem_execute (const GnomeDesktopItem *item,
 	const char *working_dir = NULL;
 	char **temp_argv = NULL;
 	int temp_argc = 0;
-	char *new_exec;
+	char *new_exec, *uris, *temp;
 	int launched = 0;
 
 	g_return_val_if_fail (item, -1);
@@ -1489,6 +1523,15 @@ ditem_execute (const GnomeDesktopItem *item,
 					  exec,
 					  args, &arg_ptr, &added_status);
 
+		if (launched == 0 && added_status == ADDED_NONE && append_uris) {
+			uris = stringify_uris (args, FALSE, FALSE);
+			temp = g_strconcat (new_exec, " ", uris, NULL);
+			g_free (uris);
+			g_free (new_exec);
+			new_exec = temp;
+			added_status = ADDED_ALL;
+		}
+
 		if (launched > 0 && added_status == ADDED_NONE) {
 			g_free (new_exec);
 			break;
@@ -1520,7 +1563,7 @@ ditem_execute (const GnomeDesktopItem *item,
 
 		if ( ! g_spawn_async (working_dir,
 				      real_argv,
-				      NULL /* envp */,
+				      envp,
 				      G_SPAWN_SEARCH_PATH /* flags */,
 				      NULL /* child_setup */,
 				      NULL /* user_data */,
@@ -1600,6 +1643,33 @@ gnome_desktop_item_launch (const GnomeDesktopItem *item,
 			   GList *file_list,
 			   GnomeDesktopItemLaunchFlags flags,
 			   GError **error)
+{
+	return gnome_desktop_item_launch_with_env (
+			item, file_list, flags, NULL, error);
+}
+
+/**
+ * gnome_desktop_item_launch_wth_env:
+ * @item: A desktop item
+ * @file_list:  Files/URIs to launch this item with, can be %NULL
+ * @flags: FIXME
+ * @envp: child's environment, or %NULL to inherit parent's
+ * @error: FIXME
+ *
+ * See gnome_desktop_item_launch for a full description. This function
+ * additionally passes an environment vector for the child process
+ * which is to be launched.
+ *
+ * Returns: The the pid of the process spawned.  If more then one
+ * process was spawned the last pid is returned.  On error -1
+ * is returned and @error is set.
+ */
+int
+gnome_desktop_item_launch_with_env (const GnomeDesktopItem       *item,
+				    GList                        *file_list,
+				    GnomeDesktopItemLaunchFlags   flags,
+				    char                        **envp,
+				    GError                      **error)
 {
 	const char *exec;
 	char *the_exec;
@@ -1662,9 +1732,10 @@ gnome_desktop_item_launch (const GnomeDesktopItem *item,
 		return -1;
 	}
 
-	ret = ditem_execute (item, the_exec, file_list,
+	ret = ditem_execute (item, the_exec, file_list, envp,
 			     (flags & GNOME_DESKTOP_ITEM_LAUNCH_ONLY_ONE),
 			     (flags & GNOME_DESKTOP_ITEM_LAUNCH_USE_CURRENT_DIR),
+			     (flags & GNOME_DESKTOP_ITEM_LAUNCH_APPEND_URIS),
 			     error);
 
 	return ret;
@@ -1869,7 +1940,7 @@ gnome_desktop_item_get_file_status (const GnomeDesktopItem *item)
 	
 	if (item->location == NULL ||
 	    gnome_vfs_get_file_info (item->location, info,
-				     GNOME_VFS_FILE_INFO_DEFAULT) != GNOME_VFS_OK) {
+				     GNOME_VFS_FILE_INFO_FOLLOW_LINKS) != GNOME_VFS_OK) {
 		gnome_vfs_file_info_unref (info);
 		return GNOME_DESKTOP_ITEM_DISAPPEARED;
 	}
@@ -2246,7 +2317,7 @@ gnome_desktop_item_set_location (GnomeDesktopItem *item, const char *location)
 			info = gnome_vfs_file_info_new ();
 
 			res = gnome_vfs_get_file_info (item->location, info,
-						       GNOME_VFS_FILE_INFO_DEFAULT);
+						       GNOME_VFS_FILE_INFO_FOLLOW_LINKS);
 
 			if (res == GNOME_VFS_OK &&
 			    info->valid_fields & GNOME_VFS_FILE_INFO_FIELDS_MTIME)
@@ -2355,6 +2426,39 @@ gnome_desktop_item_get_localestring_lang (const GnomeDesktopItem *item,
 	return lookup_locale (item, attr, language);
 }
 
+/**
+ * gnome_desktop_item_get_string_locale:
+ * @item: A desktop item
+ * @attr: An attribute name
+ *
+ * Returns the current locale that is used for the given attribute.
+ * This might not be the same for all attributes. For example, if your
+ * locale is "en_US.ISO8859-1" but attribute FOO only has "en_US" then
+ * that would be returned for attr = "FOO". If attribute BAR has
+ * "en_US.ISO8859-1" then that would be returned for "BAR".
+ *
+ * Returns: a string equal to the current locale or NULL
+ * if the attribute is invalid or there is no matching locale.
+ */
+const char *
+gnome_desktop_item_get_attr_locale (const GnomeDesktopItem *item,
+				    const char             *attr)
+{
+	const GList *list, *l;
+
+	list = gnome_i18n_get_language_list ("LC_MESSAGES");
+	for (l = list; l; l = l->next) {
+		const char *locale = l->data;
+		const char *value = NULL;
+
+		value = lookup_locale (item, attr, locale);
+		if (value)
+			return locale;
+	}
+
+	return NULL;
+}
+
 GList *
 gnome_desktop_item_get_languages (const GnomeDesktopItem *item,
 				  const char *attr)
@@ -3339,7 +3443,6 @@ ditem_load (ReadBuf *rb,
 		g_free (key);
 		key = NULL;
 	}
-	readbuf_close (rb);
 
 #undef OVERFLOW
 
@@ -3359,6 +3462,8 @@ ditem_load (ReadBuf *rb,
 	/* make sure that we set up the type */
 	setup_type (item, rb->uri);
 
+	readbuf_close (rb);
+
 	return item;
 }
 
