@@ -97,7 +97,7 @@ typedef struct _GnomeDesktopItem GnomeDesktopItem;
 #define GNOME_DESKTOP_ITEM_UNMOUNT_ICON "UnmountIcon" /* string */
 #define GNOME_DESKTOP_ITEM_SORT_ORDER	"SortOrder" /* strings */
 #define GNOME_DESKTOP_ITEM_URL		"URL" /* string */
-#define GNOME_DESKTOP_ITEM_DOC_PATH	"DocPath" /* string */
+#define GNOME_DESKTOP_ITEM_DOC_PATH	"X-GNOME-DocPath" /* string */
 
 /* The vfolder proposal */
 #define GNOME_DESKTOP_ITEM_CATEGORIES	"Categories" /* string */
@@ -114,7 +114,11 @@ typedef enum {
 	 * handle one file and we have passed many */
         GNOME_DESKTOP_ITEM_LAUNCH_ONLY_ONE = 1<<0,
 	/* Use current directory instead of home directory */
-        GNOME_DESKTOP_ITEM_LAUNCH_USE_CURRENT_DIR = 1<<1
+        GNOME_DESKTOP_ITEM_LAUNCH_USE_CURRENT_DIR = 1<<1,
+	/* Append the list of URIs to the command if no Exec
+	 * parameter is specified, instead of launching the 
+	 * app without parameters. */
+	  GNOME_DESKTOP_ITEM_LAUNCH_APPEND_URIS = 1<<2
 } GnomeDesktopItemLaunchFlags;
 
 typedef enum {
@@ -164,6 +168,12 @@ int			gnome_desktop_item_launch	     (const GnomeDesktopItem     *item,
 							      GList                      *file_list,
 							      GnomeDesktopItemLaunchFlags flags,
 							      GError                    **error);
+int			gnome_desktop_item_launch_with_env   (const GnomeDesktopItem     *item,
+							      GList                      *file_list,
+							      GnomeDesktopItemLaunchFlags flags,
+							      char                      **envp,
+							      GError                    **error);
+
 
 /* A list of files or urls dropped onto an icon This is the output
  * of gnome_vfs_uri_list_parse */
@@ -217,6 +227,9 @@ void                    gnome_desktop_item_set_string        (GnomeDesktopItem
 							      const char		 *attr,
 							      const char                 *value);
 
+const char *            gnome_desktop_item_get_attr_locale   (const GnomeDesktopItem     *item,
+							      const char		 *attr);
+
 /*
  * LocaleString type
  */
